import ctypes as T
from ..lib import lib
from .map import MapStruct, MapPointer
from .point import Point, PointStruct, PointPointer
from .block import BlockStruct, IBlock, LBlock
from ..str import state_str


class StateStruct(T.Structure):
    _fields_ = [
        ("k", PointStruct), ("p", PointStruct),
        ("block_lk", BlockStruct * 3), ("block_lp", BlockStruct * 3),
        ("block_ik", BlockStruct * 7), ("block_ip", BlockStruct * 7),
        ("map", MapStruct), ("turn", T.c_bool),
    ]


StatePointer = T.POINTER(StateStruct)

lib.State_New.argtypes = []
lib.State_New.restype = StatePointer
lib.State_Win.argtypes = []
lib.State_Win.restype = T.c_ubyte
lib.State_Delete.argtypes = [StatePointer]

lib.State_Mask.argtypes = []
lib.State_Mask.restype = StatePointer
lib.State_Copy.argtypes = []
lib.State_Copy.restype = StatePointer
lib.State_Next.argtypes = []
lib.State_Next.restype = StatePointer

lib.State_Act.argtypes = [StatePointer, T.c_ubyte, T.c_ubyte, T.c_ubyte, T.c_ubyte]
lib.State_Act.restype = StatePointer
lib.State_GetMap.argtypes = [StatePointer]
lib.State_GetMap.restype = MapPointer

lib.State_Position.argtypes = [StatePointer, T.c_bool]
lib.State_Position.restype = PointPointer


class State(object):
    def __init__(self, st=None):
        if st is not None:
            self.obj = st
        else:
            self.obj = lib.State_New()

    def __del__(self):
        lib.State_Delete(self.obj)

    def __bool__(self):
        return self.obj.contents.k.t != 0 or self.obj.contents.p.t != 0

    def act(self, action, x, y, w=0) -> 'State':
        return State(lib.State_Act(self.obj, action, x, y, w))

    @property
    def map(self):
        return lib.State_GetMap(self.obj).contents

    def __repr__(self):
        return f'State({self.obj})'

    def __str__(self):
        return f'State({self.obj})'

    def position(self, is_player=True):
        return Point(lib.State_Position(self.obj, is_player))

    @property
    def turn(self):
        return self.obj.contents.turn

    @property
    def win(self):
        return lib.State_Win(self.obj)

    def __str__(self):
        return state_str(self)

    @property
    def block_ik(self):
        return [IBlock(i) for i in self.obj.contents.block_ik]

    @property
    def block_ip(self):
        return [IBlock(i) for i in self.obj.contents.block_ip]

    @property
    def block_lk(self):
        return [LBlock(i) for i in self.obj.contents.block_lk]

    @property
    def block_lp(self):
        return [LBlock(i) for i in self.obj.contents.block_lp]

    @property
    def mask(self):
        return State(lib.State_Mask(self.obj))

    def copy(self):
        return State(lib.State_Copy(self.obj))


class colors:
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    RED = '\033[31m'
    CYAN = '\033[96m'
    GREEN = '\033[92m'
    ENDC = '\033[0m'
    PURPLE = '\033[35m'
    GRAY = '\033[90m'


def bold_sym(s: str):
    return "╋" if s == "┼" else "┃" if s == "│" else "━━━" if s == "───" else s


def state_str(self):
    if not self.turn:
        p = self.position()
        o = self.position(False)
    else:
        o = self.position()
        p = self.position(False)
    px = p.x
    py = p.y
    ox = o.x
    oy = o.y
    if px == 0 and py == 0 and ox == 0 and oy == 0:
        return "Invalid State"

    res = ''
    board = [[("┼" if y % 2 else "│") if x % 2 else ("───" if y % 2 else "   ") for x in range(17)] for y in range(17)]

    if px == ox and py == oy and px:
        board[2 * (py - 1)][2 * (px - 1)] = f'{colors.PURPLE} O {colors.GRAY}'
    else:
        if px:
            board[2 * (py - 1)][2 * (px - 1)] = f'{colors.BLUE} O {colors.GRAY}'
        if ox:
            board[2 * (oy - 1)][2 * (ox - 1)] = f'{colors.RED} O {colors.GRAY}'

    def hl(x, y):
        if x >= 0 and x < 17 and y >= 0 and y < 17:
            board[y][x] = f'{colors.ENDC}{bold_sym(board[y][x])}{colors.GRAY}'

    for x in range(10):
        for y in range(9):
            if self.map.h[y][x]:
                hl(2 * x - 1, 2 * y + 1)
                hl(2 * x - 1, 2 * y - 1)
                hl(2 * x - 1, 2 * y)

    for x in range(9):
        for y in range(10):
            if self.map.v[y][x]:
                hl(2 * x + 1, 2 * y - 1)
                hl(2 * x - 1, 2 * y - 1)
                hl(2 * x, 2 * y - 1)

    for i, line in enumerate(board):
        res = f'{colors.GRAY}  {"".join(line)}{colors.CYAN}{(i + 2) // 2 if not i % 2 else " "}{colors.GRAY}\n' + res

    res += f'   {"   ".join(map(lambda i: f"{colors.GREEN}{i}{colors.GRAY}", range(1, 10)))}{colors.ENDC}\n'
    return "%s's turn!\n" % ('POSTECH' if self.turn else 'KAIST') + res



from .types.point import Point
from .types.map import Map
from .types.block import IBlock, LBlock
from .types.state import State
from .game import Game
from .run import Run

ACTION_MOVE = 1
ACTION_PLACE_I = 2
ACTION_PLACE_L = 3


class ACTION:
    MOVE = ACTION_MOVE
    PLACE_I = ACTION_PLACE_I
    PLACE_L = ACTION_PLACE_L


BOARD_SIZE = 9
I_COUNT = 7
L_COUNT = 3



import ctypes as T
import numpy as np
from ..lib import lib


class MapStruct(T.Structure):
    _fields_ = [("h", (T.c_bool * 10) * 9), ("v", (T.c_bool * 9) * 10)]

    @property
    def list(self):
        return (
            [[self.h[y][x] for x in range(10)] for y in range(9)],
            [[self.v[y][x] for x in range(9)] for y in range(10)]
        )

    @property
    def np(self):
        li = self.list
        return (
            np.array(li[0], dtype=bool),
            np.array(li[1], dtype=bool)
        )


MapPointer = T.POINTER(MapStruct)


class Map:
    pass



import ctypes
import glob
import os

so = glob.glob(os.path.abspath(os.path.join(__file__, '../../**/*.so')), recursive=True)
if len(so) == 0:
    so = glob.glob(os.path.abspath(os.path.join(__file__, '../../**/*.lib')), recursive=True)
if len(so) == 0:
    so = glob.glob(os.path.abspath(os.path.join(__file__, '../../**/*.dll')), recursive=True)
if len(so) == 0:
    so = glob.glob(os.path.abspath(os.path.join(__file__, '../../**/*.dylib')), recursive=True)
if len(so) == 0:
    raise ImportError("Cannot import native library")

lib = ctypes.cdll.LoadLibrary(so[0])



import ctypes as T
from ..lib import lib


class PointStruct(T.Structure):
    _fields_ = [("t", T.c_ubyte)]


PointPointer = T.POINTER(PointStruct)

lib.Point_ByValue.argtypes = [T.c_int, T.c_int]
lib.Point_ByValue.restype = PointPointer
lib.Point_Copy.argtypes = [PointPointer]
lib.Point_Copy.restype = PointPointer
lib.Point_Delete.argtypes = [PointPointer]

lib.Point_X.argtypes = [PointPointer]
lib.Point_X.restype = T.c_ubyte
lib.Point_Y.argtypes = [PointPointer]
lib.Point_Y.restype = T.c_ubyte

lib.Point_Set.argtypes = [PointPointer, T.c_ubyte, T.c_ubyte]
lib.Point_SetX.argtypes = [PointPointer, T.c_ubyte]
lib.Point_SetY.argtypes = [PointPointer, T.c_ubyte]


class Point(object):
    def __init__(self, x, y=None):
        if type(x) is int:
            self.ref = lib.Point_ByValue(x, y)
        elif type(x) is PointPointer or type(x) is PointStruct:
            self.ref = x
            self.r = True

    def __eq__(self, other):
        if type(other) == Point: other = other.tuple
        return self.tuple == other

    def __del__(self):
        if 'r' in self.__dict__: return
        lib.Point_Delete(self.ref)

    def __copy__(self):
        return Point(lib.Point_Copy(self.ref))

    def __bool__(self):
        return bool(self.x or self.y)

    @property
    def x(self):
        return lib.Point_X(self.ref)

    @property
    def y(self):
        return lib.Point_Y(self.ref)

    @x.setter
    def x(self, value):
        lib.Point_SetX(self.ref, value)

    @y.setter
    def y(self, value):
        lib.Point_SetY(self.ref, value)

    def __repr__(self):
        return f'Point({self.x}, {self.y})'

    def __str__(self):
        return f'({self.x}, {self.y})'

    def set(self, x, y):
        lib.Point_Set(self.ref, x, y)

    @property
    def tuple(self):
        return (self.x, self.y)



from .game import Game


def Run(k, p, log=False):
    a, b = k, p
    s = Game()

    while True:
        if log:
            print(s)
        s.act(*a.next(s.st))
        a, b = b, a

        if s.fin:
            if log:
                print(s)
            return s.fin[0]

        if not s:
            if log:
                print(s)
            return s.fin[0]


import setuptools
import glob

cpps = glob.glob('./loq2/native/**/*.cpp', recursive=True)

native = setuptools.Extension('native', sources=cpps, extra_compile_args=['-std=c++20', '-O3'], language='c++')

install_requires = [
    'numpy==1.25.2',
    'ctypes==1.1.0',
]

setuptools.setup(
    name="loq2",
    version="0.3.7",
    author="seorii",
    author_email="me@seorii.page",
    description="League of Quoridor",
    long_description="League of Quoridor",
    long_description_content_type="text/markdown",
    url="https://github.com/seo-rii/loq2",
    packages=['loq2', 'loq2.types'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    ext_modules=[native],
    package_data={"": ["*.h"]},
)



import ctypes as T
from ..lib import lib
from .point import PointStruct, Point


class BlockStruct(T.Structure):
    _fields_ = [("p", PointStruct), ("dir", T.c_ubyte)]


BlockPointer = T.POINTER(BlockStruct)

lib.IBlock_ByValue.argtypes = [T.c_int, T.c_int, T.c_int]
lib.IBlock_ByValue.restype = BlockPointer
lib.LBlock_ByValue.argtypes = [T.c_int, T.c_int, T.c_int]
lib.LBlock_ByValue.restype = BlockPointer
lib.IBlock_Delete.argtypes = [BlockPointer]
lib.LBlock_Delete.argtypes = [BlockPointer]


class IBlock(object):
    def __init__(self, x, y=None, dir=None):
        if type(x) is BlockStruct:
            self.obj = x
            self.r = True
        elif type(x) is BlockPointer:
            self.obj = x.contents
            self.r = True
        elif y is None:
            self.obj = BlockPointer.from_address(x).contents
            self.r = True
        else:
            self.obj = lib.IBlock_ByValue(x, y, dir).contents

    def __del__(self):
        return
        if 'r' in self.__dict__: return
        lib.IBlock_Delete(self.obj)

    def __repr__(self):
        return f'IBlock({self.position}, {self.direction})'

    def __str__(self):
        return f'IBlock({self.position}, {self.direction})'

    @property
    def position(self):
        return Point(self.obj.p)

    @property
    def direction(self):
        return self.obj.dir


class LBlock(object):
    def __init__(self, x, y=None, dir=None):
        if type(x) is BlockStruct:
            self.obj = x
            self.r = True
        else:
            self.obj = lib.LBlock_ByValue(x, y, dir)

    def __del__(self):
        if 'r' in self.__dict__: return
        lib.LBlock_Delete(T.addressof(self.obj))

    def __repr__(self):
        return f'LBlock({self.position}, {self.direction})'

    def __str__(self):
        return f'LBlock({self.position}, {self.direction})'

    @property
    def position(self):
        return Point(self.obj.p)

    @property
    def direction(self):
        return self.obj.dir



from .types.state import State


class EndGameReason:
    INVALID_ACTION = 0
    WIN = 1


class Game:
    def __init__(self, state=None, acts=None):
        if acts is None:
            acts = []
        self.st = State(state)
        self.acts = acts
        self.fin = False

    def __bool__(self):
        return self.st.__bool__()

    def copy(self):
        return Game(self.st.copy(), self.acts.copy())

    def player(self, op=None):
        return self.st.turn(op)

    def act(self, action, *args):
        if self.fin:
            return False
        ret = self.st.act(action, *args)
        if not ret:
            self.fin = (1 - self.st.turn(), EndGameReason.INVALID_ACTION)
            return None
        self.st = ret
        self.acts.append((action, args))
        win = self.st.win()
        if win is not None:
            self.fin = (win, EndGameReason.WIN)
            return True
        return True

    def __str__(self):
        return self.st.__str__()
